from __future__ import annotations

from typing import List, Optional, Sequence, Union

from pydantic import BaseModel, Extra, Field, validator

from filare.models.configs import (
    CableConfig,
    ConnectorConfig,
    ConnectionConfig,
    WireConfig,
)
from filare.models.numbers import NumberAndUnit


class TemplateBaseModel(BaseModel):
    """Base class for template-facing inputs."""

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True


class TemplatePin(TemplateBaseModel):
    label: Optional[str] = None
    color: List[str] = Field(default_factory=list)
    id: Optional[str] = None

    @validator("color", pre=True)
    def _coerce_color(cls, value: Union[str, Sequence[str]]) -> List[str]:
        if value is None:
            return []
        if isinstance(value, str):
            return [value]
        return list(value)


class TemplateConnector(TemplateBaseModel):
    designator: str
    pins: List[TemplatePin] = Field(default_factory=list)
    style: Optional[str] = None
    images: List[str] = Field(default_factory=list)
    loops: List[dict] = Field(default_factory=list)

    @classmethod
    def from_config(cls, cfg: ConnectorConfig) -> "TemplateConnector":
        pins: List[TemplatePin] = []
        if cfg.pins:
            pins = [
                TemplatePin(label=p.label, color=p.color or [], id=p.id)
                for p in cfg.pins
            ]
        elif cfg.pinlabels:
            colors = cfg.pincolors or []
            for idx, label in enumerate(cfg.pinlabels):
                color = colors[idx] if idx < len(colors) else []
                pins.append(TemplatePin(label=label, color=color))
        return cls(
            designator=cfg.designator,
            pins=pins,
            style=cfg.style,
            images=cfg.images or [],
            loops=cfg.loops or [],
        )


class TemplateWire(TemplateBaseModel):
    label: Optional[str] = None
    color: List[str] = Field(default_factory=list)
    length: Optional[NumberAndUnit] = None

    @validator("color", pre=True)
    def _coerce_color(cls, value: Union[str, Sequence[str]]) -> List[str]:
        if value is None:
            return []
        if isinstance(value, str):
            return [value]
        return list(value)

    @validator("length", pre=True)
    def _coerce_length(
        cls, value: Union[str, NumberAndUnit, None]
    ) -> Optional[NumberAndUnit]:
        if value is None:
            return None
        return NumberAndUnit.to_number_and_unit(value)


class TemplateCable(TemplateBaseModel):
    designator: str
    wires: List[TemplateWire] = Field(default_factory=list)
    length: Optional[NumberAndUnit] = None
    shields: List[dict] = Field(default_factory=list)

    @classmethod
    def from_config(cls, cfg: CableConfig) -> "TemplateCable":
        wires: List[TemplateWire] = []
        if cfg.wires:
            for wire in cfg.wires:
                wires.append(
                    TemplateWire(
                        label=wire.label,
                        color=wire.color or [],
                        length=wire.length,
                    )
                )
        elif cfg.colors:
            for color in cfg.colors:
                wires.append(TemplateWire(color=color))
        return cls(
            designator=cfg.designator,
            wires=wires,
            length=NumberAndUnit.to_number_and_unit(cfg.length) if cfg.length else None,
            shields=cfg.shields or [],
        )


class TemplateConnection(TemplateBaseModel):
    endpoints: List[str]
    net: Optional[str] = None
    color: List[str] = Field(default_factory=list)

    @validator("endpoints", pre=True)
    def _coerce_endpoints(cls, value: Union[str, Sequence[str]]) -> List[str]:
        if isinstance(value, str):
            return [value]
        return list(value)

    @validator("color", pre=True)
    def _coerce_color(cls, value: Union[str, Sequence[str]]) -> List[str]:
        if value is None:
            return []
        if isinstance(value, str):
            return [value]
        return list(value)

    @classmethod
    def from_config(cls, cfg: ConnectionConfig) -> "TemplateConnection":
        return cls(endpoints=cfg.endpoints, net=cfg.net, color=cfg.color or [])



import logging
from dataclasses import asdict, dataclass, field, fields
from enum import Enum
from pathlib import Path
from typing import Dict, List, Tuple, Union

from filare.models.metadata import PagesMetadata
from filare.models.harness_quantity import HarnessQuantity
from filare.render.templates import get_template

TABLE_COLUMNS = ["sheet", "page", "quantity", "notes"]


@dataclass(frozen=True)
class IndexTableRow:
    sheet: int
    page: str
    quantity: Union[int, str] = 1
    notes: str = ""
    use_quantity: bool = True

    def get_items(self, for_pdf=False):
        if self.use_quantity:
            return (
                self.sheet,
                self.get_formatted_page(for_pdf),
                self.quantity,
                self.notes,
            )
        else:
            return (self.sheet, self.get_formatted_page(for_pdf), self.notes)

    def get_formatted_page(self, for_pdf):
        if for_pdf:
            return self.page
        return f"<a href={Path(self.page).with_suffix('.html')}>{self.page}</a>"


@dataclass(frozen=True)
class IndexTable:
    rows: List[IndexTableRow]
    header: Tuple[str]

    @staticmethod
    def use_quantity_column(metadata: PagesMetadata):
        return metadata is not None and metadata.use_qty_multipliers

    @staticmethod
    def get_index_table_header(metadata: PagesMetadata = None):
        skip = []
        if not IndexTable.use_quantity_column(metadata):
            skip.append("quantity")
        return (s.capitalize() for s in TABLE_COLUMNS if s not in skip)

    # TODO: how do we actually want to support this?
    @classmethod
    def from_pages_metadata(cls, metadata: PagesMetadata):
        header = cls.get_index_table_header(metadata)
        rows = []
        qty_multipliers = None
        if cls.use_quantity_column(metadata):
            harnesses = HarnessQuantity(
                metadata.files,
                metadata.multiplier_file_name,
                output_dir=metadata.output_dir,
            )
            harnesses.fetch_qty_multipliers_from_file()
            qty_multipliers = harnesses.multipliers

        for index, row in enumerate(metadata.output_names):
            if str(row) == "titlepage":
                rows.append(
                    IndexTableRow(
                        sheet=1,
                        page=metadata.titlepage,
                        quantity="",
                        notes="",
                        use_quantity=cls.use_quantity_column(metadata),
                    )
                )
                continue
            quantity = qty_multipliers[row] if qty_multipliers is not None else 1
            rows.append(
                IndexTableRow(
                    sheet=index + 1,
                    page=row,
                    quantity=quantity,
                    notes=metadata.pages_notes.get(row, ""),
                    use_quantity=cls.use_quantity_column(metadata),
                )
            )
        return cls(rows=rows, header=header)

    def render(self, options):
        return get_template("index_table.html").render(
            {
                "index_table": self,
                "options": options,
            }
        )



from dataclasses import dataclass, field

from filare.models.utils import html_line_breaks


@dataclass
class MultilineHypertext:
    raw: str  # Hypertext possibly also including newlines to break lines in diagram output

    @staticmethod
    def to(value):
        if isinstance(value, MultilineHypertext):
            return value

        if isinstance(value, str):
            return MultilineHypertext(raw=value)

        if isinstance(value, list):
            return MultilineHypertext(raw="<br>".join(value))

        return MultilineHypertext("")

    @property
    def clean(self):
        return html_line_breaks(self.raw)

    def __repr__(self):
        return self.raw

    def is_empty(self):
        return not self.raw



"""Domain models for connectors, cables, BOM, metadata, and related helpers."""

from filare.models.bom import *  # noqa: F401,F403
from filare.models.colors import *  # noqa: F401,F403
from filare.models.cable import *  # noqa: F401,F403
from filare.models.component import *  # noqa: F401,F403
from filare.models.connector import *  # noqa: F401,F403
from filare.models.configs import *  # noqa: F401,F403
from filare.models.dataclasses import *  # noqa: F401,F403
from filare.models.harness import *  # noqa: F401,F403
from filare.models.harness_quantity import *  # noqa: F401,F403
from filare.models.metadata import *  # noqa: F401,F403
from filare.models.options import *  # noqa: F401,F403
from filare.models.types import *  # noqa: F401,F403
from filare.models.template_inputs import *  # noqa: F401,F403



# -*- coding: utf-8 -*-

import re
from pathlib import Path
from typing import List

awg_equiv_table = {
    "0.09": "28",
    "0.14": "26",
    "0.25": "24",
    "0.34": "22",
    "0.5": "21",
    "0.75": "20",
    "1": "18",
    "1.5": "16",
    "2.5": "14",
    "4": "12",
    "6": "10",
    "10": "8",
    "16": "6",
    "25": "4",
    "35": "2",
    "50": "1",
}

mm2_equiv_table = {v: k for k, v in awg_equiv_table.items()}


def awg_equiv(mm2):
    return awg_equiv_table.get(str(mm2), "Unknown")


def mm2_equiv(awg):
    return mm2_equiv_table.get(str(awg), "Unknown")


def expand(yaml_data):
    # yaml_data can be:
    # - a singleton (normally str or int)
    # - a list of str or int
    # if str is of the format '#-#', it is treated as a range (inclusive) and expanded
    output = []
    if not isinstance(yaml_data, list):
        yaml_data = [yaml_data]
    for e in yaml_data:
        e = str(e)
        if "-" in e:
            a, b = e.split("-", 1)
            try:
                a = int(a)
                b = int(b)
                if a < b:
                    for x in range(a, b + 1):
                        output.append(x)  # ascending range
                elif a > b:
                    for x in range(a, b - 1, -1):
                        output.append(x)  # descending range
                else:  # a == b
                    output.append(a)  # range of length 1
            except:
                # '-' was not a delimiter between two ints, pass e through unchanged
                output.append(e)
        else:
            try:
                x = int(e)  # single int
            except Exception:
                x = e  # string
            output.append(x)
    return output


def get_single_key_and_value(d: dict):
    # used for defining a line in a harness' connection set
    # E.g. for the YAML input `- X1: 1`
    # this function returns a tuple in the form ("X1", "1")
    return next(iter(d.items()))


def int2tuple(inp):
    if isinstance(inp, tuple):
        output = inp
    else:
        output = (inp,)
    return output


def flatten2d(inp):
    return [
        [str(item) if not isinstance(item, List) else ", ".join(item) for item in row]
        for row in inp
    ]


# TODO: move to hyperlink
def html_line_breaks(inp):
    return remove_links(inp).replace("\n", "<br />") if isinstance(inp, str) else inp


# TODO: move to hyperlink
def remove_links(inp):
    return (
        re.sub(r"<[aA] [^>]*>([^<]*)</[aA]>", r"\1", inp)
        if isinstance(inp, str)
        else inp
    )


def clean_whitespace(inp):
    return " ".join(inp.split()).replace(" ,", ",") if isinstance(inp, str) else inp


def smart_file_resolve(filename: Path, possible_paths: (Path, List[Path])) -> Path:
    if isinstance(possible_paths, Path) or isinstance(possible_paths, str):
        possible_paths = [possible_paths]
    if filename.is_absolute():
        if filename.exists():
            return filename
        else:
            raise Exception(f"{filename} does not exist.")
    else:  # search all possible paths in decreasing order of precedence
        for path in possible_paths:
            combined_path = (path / filename).resolve()
            if combined_path.exists():
                return combined_path
        raise Exception(
            f"{filename} was not found in any of the following locations: \n"
            + "\n".join(str(p) for p in possible_paths)
        )



"""Higher-level parsing helpers for harness YAML files."""

from pathlib import Path
from typing import Iterable, Tuple

from filare.parser.yaml_loader import parse_concat_merge_files, parse_merge_files


def parse_harness_files(
    component_files: Tuple[Path, ...],
    harness_files: Tuple[Path, ...],
    metadata_files: Tuple[Path, ...],
):
    """Parse component + harness YAML into a merged dict."""
    return parse_concat_merge_files(component_files + harness_files, metadata_files)


def parse_metadata_files(metadata_files: Iterable[Path]):
    """Parse metadata-only YAML files."""
    if not metadata_files:
        return {}
    return parse_merge_files(tuple(metadata_files))



# -*- coding: utf-8 -*-
"""
Shim module for rendering helpers.

Functions are implemented in:
- render.assets: embed_svg_images, embed_svg_images_file, get_mime_subtype
- render.pdf: generate_pdf_output
- render.html: generate_html_output, generate_shared_bom, generate_titlepage
"""

from filare.render.assets import (
    embed_svg_images,
    embed_svg_images_file,
    get_mime_subtype,
)
from filare.render.html import (
    generate_html_output,
    generate_shared_bom,
    generate_titlepage,
)
from filare.render.pdf import generate_pdf_output

__all__ = [
    "embed_svg_images",
    "embed_svg_images_file",
    "get_mime_subtype",
    "generate_html_output",
    "generate_shared_bom",
    "generate_titlepage",
    "generate_pdf_output",
]



from typing import List, Optional

from pydantic import BaseModel, Field, root_validator, validator

from filare.models.colors import ColorOutputMode, SingleColor
from filare.models.types import PlainText


class PageOptions(BaseModel):
    # formatting toggles
    show_bom: bool = True
    bom_updated_position: str = ""
    show_index_table: bool = True
    index_table_on_right: bool = True
    index_table_updated_position: str = ""
    show_notes: bool = True
    notes_on_right: bool = True
    notes_width: str = "100mm"

    # diagram colors
    bgcolor: SingleColor = Field(default_factory=lambda: SingleColor(inp="WH"))
    bgcolor_node: SingleColor = Field(default_factory=lambda: SingleColor(inp="WH"))
    bgcolor_connector: Optional[SingleColor] = None
    bgcolor_cable: Optional[SingleColor] = None
    bgcolor_bundle: Optional[SingleColor] = None
    color_output_mode: ColorOutputMode = ColorOutputMode.EN_UPPER

    # dimensions
    bom_rows: int = 0
    titleblock_rows: int = 9
    bom_row_height: float = 4.25
    titleblock_row_height: float = 4.25
    index_table_row_height: float = 4.25

    # misc
    fontname: PlainText = "arial"
    mini_bom_mode: bool = True
    template_separator: str = "."
    for_pdf: bool = False
    _pad: int = 0
    _template_paths: List = Field(default_factory=list)
    _image_paths: List = Field(default_factory=list)

    @validator(
        "bgcolor",
        "bgcolor_node",
        "bgcolor_connector",
        "bgcolor_cable",
        "bgcolor_bundle",
        pre=True,
    )
    def _to_single_color(cls, value):
        return value if isinstance(value, SingleColor) else SingleColor(inp=value)

    @root_validator
    def _fill_missing_colors(cls, values):
        values["bgcolor_node"] = values.get("bgcolor_node") or values.get("bgcolor")
        values["bgcolor_connector"] = values.get("bgcolor_connector") or values.get(
            "bgcolor_node"
        )
        values["bgcolor_cable"] = values.get("bgcolor_cable") or values.get(
            "bgcolor_node"
        )
        values["bgcolor_bundle"] = values.get("bgcolor_bundle") or values.get(
            "bgcolor_cable"
        )
        return values

    @validator(
        "bom_rows",
        "titleblock_rows",
        pre=True,
    )
    def _coerce_int(cls, value):
        if isinstance(value, list):
            return len(value)
        return int(value)

    @validator(
        "bom_row_height",
        "titleblock_row_height",
        "index_table_row_height",
        pre=True,
    )
    def _coerce_float(cls, value):
        return float(value)

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        underscore_attrs_are_private = True


def get_page_options(parsed_data, page_name: str):
    """Get the page options

    uses: the page\'s options   -> general options -> default options
        ('{page_name}_options') ->    ('options')  -> {}
    """
    page_options_name = f"{page_name}_options"
    if page_options_name in parsed_data:
        return PageOptions(**parsed_data[page_options_name])
    return PageOptions(**parsed_data.get("options", {}))



from pathlib import Path

import jinja2


def get_template(template_name, extension=""):
    templates_root = Path(__file__).resolve().parent.parent / "templates"
    template_file_path = jinja2.FileSystemLoader(templates_root)
    jinja_env = jinja2.Environment(
        loader=template_file_path, undefined=jinja2.StrictUndefined
    )

    return jinja_env.get_template(template_name + extension)



import json
import logging
from pathlib import Path
from typing import Dict, List, Optional

import click
from pydantic import BaseModel, Field, root_validator


class HarnessQuantity(BaseModel):
    harnesses: List[Path]
    multiplier_file_name: str = "quantity_multipliers.txt"
    output_dir: Optional[Path] = None
    multipliers: Dict[str, int] = Field(default_factory=dict)
    folder: Path
    qty_multipliers: Path

    def __init__(
        self,
        harnesses,
        multiplier_file_name="quantity_multipliers.txt",
        output_dir=None,
        **data,
    ):
        super().__init__(
            harnesses=list(harnesses),
            multiplier_file_name=multiplier_file_name,
            output_dir=output_dir,
            **data,
        )

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = True

    @root_validator(pre=True)
    def _derive_paths(cls, values):
        harnesses = values.get("harnesses") or []
        if not harnesses:
            return values
        folder = values.get("output_dir") or harnesses[0].parent
        values["folder"] = folder
        values["qty_multipliers"] = folder / values.get(
            "multiplier_file_name", "quantity_multipliers.txt"
        )
        return values

    @property
    def harness_names(self):
        return [harness.stem for harness in self.harnesses]

    def __getitem__(self, harness):
        return self.multipliers[harness]

    def fetch_qty_multipliers_from_file(self):
        if self.qty_multipliers.is_file():
            with open(self.qty_multipliers, "r") as f:
                try:
                    self.multipliers = json.load(f)
                except json.decoder.JSONDecodeError as err:
                    raise ValueError(
                        f"Invalid format for file {self.qty_multipliers}, error: {err}"
                    )
        else:
            self.get_qty_multipliers_from_user()
            self.save_qty_multipliers_to_file()
        self.check_all_multipliers_defined()

    def check_all_multipliers_defined(self):
        for name in self.harness_names:
            assert (
                name in self.multipliers
            ), f"No multiplier defined for harness {name}, maybe delete the multiplier_file {self.qty_multipliers}"

    def get_qty_multipliers_from_user(self):
        for name in self.harness_names:
            try:
                self.multipliers[name] = int(
                    input("Quantity multiplier for {}? ".format(name))
                )
            except ValueError:
                logging.warning("Quantity multiplier must be an integer!")
                break

    def save_qty_multipliers_to_file(self):
        with open(self.qty_multipliers, "w") as f:
            json.dump(self.multipliers, f)

    def retrieve_harness_qty_multiplier(self, bom_file):
        return int(self[Path(Path(bom_file).stem).stem])


@click.command(no_args_is_help=True)
@click.argument(
    "files",
    type=click.Path(
        exists=True,
        readable=True,
        dir_okay=False,
        path_type=Path,
    ),
    nargs=-1,
    required=True,
)
@click.option(
    "-m",
    "--multiplier-file-name",
    default="quantity_multipliers.txt",
    type=str,
    help="name of file used to fetch or save the qty_multipliers",
)
@click.option(
    "-f",
    "--force-new",
    is_flag=True,
    type=bool,
    help="if set, will always ask for new multipliers",
)
def qty_multipliers(files, multiplier_file_name, force_new):
    harnesses = HarnessQuantity(files, multiplier_file_name)
    if force_new:
        harnesses.qty_multipliers.unlink(missing_ok=True)

    harnesses.fetch_qty_multipliers_from_file()
    qty_multipliers = harnesses.multipliers
    return



# -*- coding: utf-8 -*-
import logging
from dataclasses import asdict, dataclass, field, fields
from enum import Enum, IntEnum
from itertools import zip_longest
from typing import Any, Dict, List, Optional, Tuple, Union

from filare.models.hypertext import MultilineHypertext
from filare.models.image import Image
from filare.models.numbers import NumberAndUnit
from filare.models.partnumber import PartNumberInfo, PartnumberInfoList
from filare.models.bom import BomEntry, BomEntryBase
from filare.models.colors import (
    COLOR_CODES,
    ColorOutputMode,
    MultiColor,
    SingleColor,
    get_color_by_colorcode_index,
)
from filare.models.utils import awg_equiv, mm2_equiv, remove_links

# Each type alias have their legal values described in comments
# - validation might be implemented in the future
PlainText = str  # Text not containing HTML tags nor newlines

Designator = PlainText  # Case insensitive unique name of connector or cable

# Type combinations
Pin = Union[int, PlainText]  # Pin identifier
PinIndex = int  # Zero-based pin index
Wire = Union[int, PlainText]  # Wire number or Literal['s'] for shield
NoneOrMorePins = Union[
    Pin, Tuple[Pin, ...], None
]  # None, one, or a tuple of pin identifiers
NoneOrMorePinIndices = Union[
    PinIndex, Tuple[PinIndex, ...], None
]  # None, one, or a tuple of zero-based pin indices
OneOrMoreWires = Union[Wire, Tuple[Wire, ...]]  # One or a tuple of wires


Side = Enum("Side", "LEFT RIGHT")

AUTOGENERATED_PREFIX = "AUTOGENERATED_"

BomCategory = IntEnum(  # to enforce ordering in BOM
    "BomEntry", "CONNECTOR CABLE WIRE PIN ADDITIONAL BUNDLE"
)
QtyMultiplierConnector = Enum(
    "QtyMultiplierConnector", "PINCOUNT POPULATED CONNECTIONS"
)
QtyMultiplierCable = Enum(
    "QtyMultiplierCable", "WIRECOUNT TERMINATION LENGTH TOTAL_LENGTH"
)


@dataclass
class PinClass:
    index: int
    id: str
    label: str
    color: MultiColor
    parent: str  # designator of parent connector
    _num_connections = 0  # incremented in Connector.connect()
    _anonymous: bool = False  # true for pins on autogenerated connectors
    _simple: bool = False  # true for simple connector

    # TODO: support a "crimp" defined by parent

    def __post_init__(self):
        if self.label and "__" in self.label:
            self.label, pin = self.label.split("__")
            self.index = int(pin) - 1

    def __str__(self):
        snippets = [  # use str() for each in case they are int or other non-str
            str(self.parent) if not self._anonymous else "",
            str(self.id) if not self._anonymous and not self._simple else "",
            str(self.label) if self.label else "",
        ]
        return ":".join([snip for snip in snippets if snip != ""])

    @property
    def pin(self):
        return self.index + 1

    @property
    def category(self):
        return BomCategory.PIN


@dataclass
class Component:
    category: Optional[str] = None  # currently only used by cables, to define bundles
    type: Union[MultilineHypertext, List[MultilineHypertext]] = None
    subtype: Union[MultilineHypertext, List[MultilineHypertext]] = None

    # the following are provided for user convenience and should not be accessed later.
    # their contents are loaded into partnumbers during the child class __post_init__()
    pn: str = None
    manufacturer: str = None
    mpn: str = None
    supplier: str = None
    spn: str = None

    # BOM info
    qty: NumberAndUnit = field(
        default_factory=lambda: NumberAndUnit(number=1, unit=None)
    )
    amount: Optional[NumberAndUnit] = None
    ignore_in_bom: bool = False
    id: Optional[str] = None  # to be filled after harness is built
    designators: [List] = field(
        default_factory=list
    )  # Used only for additional components

    # Utility
    parent: Optional = None
    additional_components: Optional[List] = field(default_factory=list)
    qty_multiplier: Union[QtyMultiplierConnector, QtyMultiplierCable, int] = 1
    _qty_multiplier_computed: Union[int, float] = 1

    # style
    bgcolor: SingleColor = None

    def __hash__(self):
        """Provide a hash for this component dataclass.

        Any component using same part should have the same hash
        """
        return hash(self.partnumbers)

    def __str__(self) -> str:
        return f"{self.type}{', ' + self.subtype.raw if self.subtype.raw else ''}"

    def __post_init__(self):
        self.qty = NumberAndUnit.to_number_and_unit(self.qty)
        self.amount = NumberAndUnit.to_number_and_unit(self.amount)

        self.type = MultilineHypertext.to(self.type)
        self.subtype = MultilineHypertext.to(self.subtype)

        if self.qty_multiplier:
            try:
                self.qty_multiplier = int(self.qty_multiplier)
            except:
                pass

        if isinstance(self.pn, list):
            raise RuntimeError(f"PN ({self.pn}) should not be a list")

        for i, item in enumerate(self.additional_components):
            if isinstance(item, Component):
                continue
            elif isinstance(item, dict):
                self.additional_components[i] = Component(
                    **item, category=BomCategory.ADDITIONAL, parent=self
                )
            else:
                raise ValueError(
                    f"additional component {item} should be a Component or a dict, is {type(item)}"
                )

        if self.category is None:
            raise RuntimeError(f"category should be defined for {self}")

    def compute_qty_multipliers(self):
        pass

    @property
    def bom_entry(self):
        self.compute_qty_multipliers()
        return BomEntryBase(
            qty=self.qty,
            partnumbers=self.partnumbers,
            amount=self.amount,
            qty_multiplier=self._qty_multiplier_computed,
            description=str(self),
            category=self.category,
            designators=self.designators,
            ignore_in_bom=self.ignore_in_bom,
        )

    @property
    def partnumbers(self):
        return PartNumberInfo(
            pn=self.pn,
            manufacturer=self.manufacturer,
            mpn=self.mpn,
            supplier=self.supplier,
            spn=self.spn,
        )


@dataclass
class GraphicalComponent(Component):  # abstract class
    # component properties
    designator: Designator = ""
    color: Optional[MultiColor] = None
    image: Optional[Image] = None
    additional_components: List[Component] = field(default_factory=list)
    notes: Optional[MultilineHypertext] = None
    # BOM options
    add_up_in_bom: Optional[bool] = None
    # rendering options
    bgcolor: Optional[SingleColor] = None
    bgcolor_title: Optional[SingleColor] = None
    show_name: Optional[bool] = None

    def __hash__(self):
        return hash(super())

    def __post_init__(self):
        super().__post_init__()
        self.notes = MultilineHypertext.to(self.notes)

        self.designator = remove_links(self.designator)


@dataclass
class Loop:
    first: PinClass = None
    second: PinClass = None
    side: Side = None
    show_label: bool = True
    color: Optional[MultiColor] = None

    def __post_init__(self):
        if self.side is not None:
            if isinstance(self.side, str):
                side_str = self.side.upper()
            else:
                side_str = str(self.side).upper()

            if side_str == "LEFT":
                self.side = Side.LEFT
            elif side_str == "RIGHT":
                self.side = Side.RIGHT
            else:
                self.side = None

        self.color = MultiColor(self.color)

    @property
    def label(self):
        if self.show_label:
            return f"{self.first}⇿{self.second}"
        else:
            return ""

    def html_color(self):
        return self.color[0].html


@dataclass
class Connector(GraphicalComponent):
    # connector-specific properties
    style: Optional[str] = None
    loops: List[Loop] = field(default_factory=list)
    # pin information in particular
    pincount: Optional[int] = None
    pins: List[Pin] = field(default_factory=list)  # legacy
    pinlabels: List[Pin] = field(default_factory=list)  # legacy
    pincolors: List[str] = field(default_factory=list)  # legacy
    pin_objects: Dict[Any, PinClass] = field(default_factory=dict)  # new
    # rendering option
    show_pincount: Optional[bool] = None
    hide_disconnected_pins: bool = False

    ports_left: bool = True
    ports_right: bool = False

    def __hash__(self):
        return hash(super())

    def __str__(self) -> str:
        substrs = [
            "Connector",
            self.type,
            self.subtype,
            f"{self.pincount} pins" if self.show_pincount else None,
            str(self.color) if self.color else None,
        ]
        return ", ".join([str(s) for s in substrs if s is not None and s != ""])

    @property
    def is_autogenerated(self):
        return self.designator.startswith(AUTOGENERATED_PREFIX)

    @property
    def has_pincolors(self):
        return any(_pin.color for _pin in self.pin_objects.values())

    def should_show_pin(self, pin_id):
        return (
            not self.hide_disconnected_pins
            or self.pin_objects[pin_id]._num_connections > 0
        )

    def pins_to_show(self):
        return [p for k, p in self.pin_objects.items() if self.should_show_pin(k)]

    def __post_init__(self) -> None:
        self.category = BomCategory.CONNECTOR
        super().__post_init__()

        self.bgcolor = SingleColor(self.bgcolor)
        self.bgcolor_title = SingleColor(self.bgcolor_title)
        self.color = MultiColor(self.color)

        if isinstance(self.image, dict):
            self.image = Image(**self.image)

        self._ports_left_set = False
        self._ports_right_set = False

        if self.pincount is not None:
            self.pincount = int(self.pincount)

        if self.style == "simple":
            if self.pincount and self.pincount > 1:
                raise Exception(
                    "Connectors with style set to simple may only have one pin"
                )
            self.pincount = 1

        if self.pincount is None:
            self.pincount = max(
                len(self.pins), len(self.pinlabels), len(self.pincolors)
            )
            if not self.pincount:
                raise Exception(
                    "You need to specify at least one: "
                    "pincount, pins, pinlabels, or pincolors"
                )

        # create default list for pins (sequential) if not specified
        if not self.pins:
            self.pins = list(range(1, self.pincount + 1))

        if len(self.pins) != len(set(self.pins)):
            raise Exception("Pins are not unique")

        # convert all pins which can be to int
        def to_int_pin(pin):
            try:
                return int(pin)
            except ValueError:
                return pin

        self.pins = [to_int_pin(p) for p in self.pins]

        # all checks have passed
        pin_tuples = zip_longest(
            self.pins,
            self.pinlabels,
            self.pincolors,
        )
        for pin_index, (pin_id, pin_label, pin_color) in enumerate(pin_tuples):
            self.pin_objects[pin_id] = PinClass(
                index=pin_index,
                id=pin_id,
                label=pin_label,
                color=MultiColor(pin_color),
                parent=self.designator,
                _anonymous=self.is_autogenerated,
                _simple=self.style == "simple",
            )

        if self.show_name is None:
            self.show_name = self.style != "simple" and not self.is_autogenerated

        if self.show_pincount is None:
            # hide pincount for simple (1 pin) connectors by default
            self.show_pincount = self.style != "simple"

        def get_pin_object(value):
            # is label?
            pin_match_label = [p for p in self.pin_objects.values() if p.label == value]
            if pin_match_label:
                return pin_match_label[0]
            err = f"{value} not found in {self.pinlabels}"

            try:
                value = int(value)
            except ValueError as exc:
                raise ValueError(f"{err} and is not an int")

            pin_match_id = [p for p in self.pin_objects.values() if p.id == value]
            if not pin_match_id:
                raise ValueError(f"{err} and is not one of the pins {self.pins}")
            return pin_match_id[0]

        normalized_loops = []
        for loop in self.loops:
            if isinstance(loop, (list, tuple)) and len(loop) >= 2:
                normalized_loops.append(
                    {
                        "first": loop[0],
                        "second": loop[1],
                        "side": None,
                        "show_label": True,
                        "color": None,
                    }
                )
            elif isinstance(loop, dict):
                normalized_loops.append(loop)
            else:
                raise ValueError(f"Unsupported loop definition: {loop}")

        self.loops = [
            Loop(
                first=get_pin_object(loop["first"]),
                second=get_pin_object(loop["second"]),
                side=loop.get("side"),
                show_label=loop.get("show_label", True),
                color=loop.get("color"),
            )
            for loop in normalized_loops
        ]

        for loop in self.loops:
            self.activate_pin(loop.first.id, side=loop.side, is_connection=True)
            self.activate_pin(loop.second.id, side=loop.side, is_connection=True)

    def activate_pin(self, pin_id, side: Side = None, is_connection=True) -> None:
        if is_connection:
            self.pin_objects[pin_id]._num_connections += 1
        if side == Side.LEFT:
            self._ports_left_set = True
            self.ports_left = True
            if not self._ports_right_set:
                self.ports_right = False
        elif side == Side.RIGHT:
            self._ports_right_set = True
            self.ports_right = True
            if not self._ports_left_set:
                self.ports_left = False

    def compute_qty_multipliers(self):
        # do not run before all connections in harness have been made!
        num_populated_pins = len(
            [pin for pin in self.pin_objects.values() if pin._num_connections > 0]
        )
        num_connections = sum(
            [pin._num_connections for pin in self.pin_objects.values()]
        )
        qty_multipliers_computed = {
            "PINCOUNT": self.pincount,
            "POPULATED": num_populated_pins,
            "CONNECTIONS": num_connections,
        }
        for subitem in self.additional_components:
            if isinstance(subitem.qty_multiplier, str):
                computed_factor = qty_multipliers_computed[
                    subitem.qty_multiplier.upper()
                ]
            # if isinstance(subitem.qty_multiplier, QtyMultiplierConnector):
            #    computed_factor = qty_multipliers_computed[subitem.qty_multiplier.name.upper()]
            # elif isinstance(subitem.qty_multiplier, QtyMultiplierCable):
            #    raise Exception("Used a cable multiplier in a connector!")
            elif isinstance(subitem.qty_multiplier, int) or isinstance(
                subitem.qty_multiplier, float
            ):
                computed_factor = subitem.qty_multiplier
            else:
                raise ValueError(
                    f'Unexpected qty multiplier "{subitem.qty_multiplier}"'
                )
            subitem._qty_multiplier_computed = computed_factor


@dataclass
class WireClass(GraphicalComponent):
    parent: str = None  # designator of parent cable/bundle
    # wire-specific properties
    index: int = None
    label: str = ""
    color: MultiColor = None
    # inheritable from parent cable
    gauge: Optional[NumberAndUnit] = None
    length: Optional[NumberAndUnit] = None
    ignore_in_bom: Optional[bool] = False

    is_shield = False

    def __hash__(self):
        return hash((self.partnumbers, self.gauge_str, str(self.color)))

    def __str__(self) -> str:
        substrs = [
            "Wire",
            self.type.raw,
            self.subtype.raw,
            self.gauge_str,
            str(self.color) if self.color else None,
        ]
        desc = ", ".join([s for s in substrs if s is not None and s != ""])
        return desc

    def __post_init__(self):
        self.category = BomCategory.WIRE
        super().__post_init__()

    def wireinfo(self, parent_is_bundle=False):
        wireinfo = []
        if not parent_is_bundle and not self.is_shield:
            wireinfo.append(self.id)
        if self.color:
            wireinfo.append(str(self.color))
        if self.label:
            wireinfo.append(self.label)
        return ":".join(wireinfo)

    @property
    def port(self):
        return f"w{self.index+1}"

    @property
    def partnumbers(self):
        _partnumbers = super().partnumbers
        if not _partnumbers.mpn and self.color is not None:
            _partnumbers.mpn = self.get_mpn_if_belden()
        return _partnumbers

    @property
    def bom_entry(self):
        self.compute_qty_multipliers()
        return BomEntry(
            qty=self.length,
            partnumbers=self.partnumbers,
            id=self.id,
            description=str(self),
            category=self.category,
            ignore_in_bom=self.ignore_in_bom,
        )

    @property
    def gauge_str(self):
        if not self.gauge:
            return None
        number = (
            int(self.gauge.number) if self.gauge.unit == "AWG" else self.gauge.number
        )
        actual_gauge = f"{number} {self.gauge.unit}"
        actual_gauge = actual_gauge.replace("mm2", "mm\u00b2")
        return actual_gauge

    @property
    def gauge_str_with_equiv(self):
        if not self.gauge:
            return None
        actual_gauge = self.gauge_str
        equivalent_gauge = ""
        if self.show_equiv:
            # convert unit if known
            if self.gauge.unit == "mm2":
                equivalent_gauge = f" ({awg_equiv(self.gauge.number)} AWG)"
            elif self.gauge.unit.upper() == "AWG":
                equivalent_gauge = f" ({mm2_equiv(self.gauge.number)} mm2)"
        out = f"{actual_gauge}{equivalent_gauge}"
        out = out.replace("mm2", "mm\u00b2")
        return out

    @property
    def length_str(self):
        if not self.length:
            return None
        return str(self.length)

    belden_color = {
        "BN": "001",
        "RD": "002",
        "OG": "003",
        "YE": "004",
        "GN": "005",
        "TQ": "006",  # (For Belden: light blue. For Filare: turquoise)
        "VT": "007",
        "GY": "008",
        "WH": "009",
        "BK": "010",
        "BG": "011",
        "PK": "012",
        "BU": "013",
        "BKRD": "015",  # (for Belden: white/red)
        "BKGN": "016",  # (for Belden: white/green)
        "BKYE": "017",  # (for Belden: white/yellow)
        "BKBU": "018",  # (for Belden: white/blue)
        "BKBN": "019",  # (for Belden: white/brown)
        "BKOG": "020",  # (for Belden: white/orange)
        "BKGY": "021",  # (for Belden: white/gray)
        "BKVT": "022",  # (for Belden: white/purple)
        # (1) Why use BKRD instead of WHRD, since Belden only sells white/red?
        #    - WHRD is impractical to use in Wireviz (white wire sides on white background does not help with identification)
        #    - BKRD is clearly distinguishable, and comes handy to use in Wireviz, as the representation of a GND rail associated (even twisted, if applicable) with a specific RD signal wire.
        # (2) For all wire colors see:
        # https://www.belden.com/dfsmedia/f1e38517e0cd4caa8b1acb6619890f5e/7806-source/options/view/cabling-solutions-for-industrial-applications-catalog-belden-09-2020#page=153
    }
    belden_tfe_base_mpn = {
        # Leftmost in list is the prefered MPN
        # NOTE (lal 2022-12-20):  this prefered MPN is arbitrary ATM
        "16 AWG": ["83030", "83010"],
        "18 AWG": ["83029", "83009"],
        "20 AWG": ["83028", "83027", "83007", "83008"],
        "22 AWG": ["83025", "83026", "83005", "83006", "83049", "83050"],
        "24 AWG": ["83003", "83004", "83023", "83047", "83048"],
        "26 AWG": ["83002", "83046"],
        "28 AWG": ["83001", "83045"],
        "30 AWG": ["83000", "83043"],
        "32 AWG": ["83041"],
        # see: https://www.belden.com/dfsmedia/f1e38517e0cd4caa8b1acb6619890f5e/7806-source/options/view/cabling-solutions-for-industrial-applications-catalog-belden-09-2020#page=136
    }

    @property
    def is_belden(self):
        if "belden" in self.manufacturer.lower():
            return True
        return False

    def get_belden_color(self, color):
        if color not in self.belden_color:
            logging.warn(
                f"{self}: Color '{self.color}' not found in belden colors {list(self.belden_color.keys())}, defaulting to BK"
            )
            return self.belden_color["BK"]
        return self.belden_color[color]

    def gen_belden_cable_with_alternate(self):
        # Gauge and mpn base
        try:
            parts = self.belden_tfe_base_mpn[self.gauge_str]
        except KeyError:
            raise ValueError(
                f"Couldn't find a belden TFE wire for wire of {self.gauge_str}"
            )

        color = self.get_belden_color(str(self.color))

        if not color:
            raise ValueError(f"Failed to find a color for property: {self}")

        # Create the list of mpn
        roll_length = 100
        mpn_list = [f"{mpn} {color}{roll_length}" for mpn in parts]

        main_part = mpn_list[0]
        alternates = mpn_list[1:] if len(mpn_list) > 1 else []
        return (main_part, alternates)

    # TODO: clean me up
    def get_mpn_if_belden(self):
        if self.manufacturer and not self.mpn:
            if self.is_belden:
                main_part, alternates = self.gen_belden_cable_with_alternate()
                return main_part
                if alternates:
                    logging.info(
                        f'Alternate part{"s" if len(alternates) > 1 else ""} available for {self.gauge_str}, color {self.color}: {alternates}'
                    )
            else:
                logging.info(
                    f'Not updating part for manufacturer {self.manufacturer}, only "belden" supported'
                )
        else:
            logging.info(f"Not updating part, no manufacturer provided")
        return self.mpn if self.mpn else ""


@dataclass
class ShieldClass(WireClass):
    is_shield = True
    pass  # TODO, for wires with multiple shields more shield details, ...

    def __hash__(self):
        return hash(self.partnumbers)


@dataclass
class Connection:
    from_: PinClass = None
    via: Union[WireClass, ShieldClass] = None
    to: PinClass = None


@dataclass
class Cable(WireClass):
    # cable-specific properties
    gauge: Optional[NumberAndUnit] = None
    length: Optional[NumberAndUnit] = None
    color_code: Optional[str] = None
    # wire information in particular
    wirecount: Optional[int] = None
    shield: Union[bool, MultiColor] = False
    colors: List[str] = field(default_factory=list)  # legacy
    wirelabels: List[Wire] = field(default_factory=list)  # legacy
    wire_objects: Dict[Any, WireClass] = field(default_factory=dict)  # new
    # internal
    _connections: List[Connection] = field(default_factory=list)
    # rendering options
    show_name: Optional[bool] = None
    show_equiv: bool = False
    show_wirecount: bool = True

    def __hash__(self):
        if self.is_bundle:
            return hash(tuple([hash(w) for w in self.wire_objects.values()]))
        else:
            return hash(super())

    def __str__(self) -> str:
        substrs = [
            ("", "Cable"),
            (", ", self.type),
            (", ", self.subtype),
            (", ", self.wirecount),
            (" ", f"x {self.gauge_str}" if self.gauge else "wires"),
            (" ", "shielded" if self.shield else None),
            (", ", str(self.color) if self.color else None),
        ]
        if self.is_bundle:
            substrs += [
                (f"\n\t{i}: ", w) for i, w in enumerate(self.wire_objects.values())
            ]
        desc = "".join(
            [f"{s[0]}{s[1]}" for s in substrs if s[1] is not None and s[1] != ""]
        )
        return desc

    @property
    def partnumbers(self):
        if self.is_bundle:
            return PartnumberInfoList(
                pn_list=[w.partnumbers for w in self.wire_objects.values()]
            )
        else:
            return super().partnumbers

    @property
    def bom_entry(self):
        self.compute_qty_multipliers()
        if self.is_bundle:
            return [w.bom_entry for w in self.wire_objects.values()]
        else:
            return BomEntryBase(
                qty=self.qty,
                partnumbers=self.partnumbers,
                amount=self.amount,
                qty_multiplier=self._qty_multiplier_computed,
                description=str(self),
                category=self.category,
                designators=self.designators,
                ignore_in_bom=self.ignore_in_bom,
            )

    @property
    def is_bundle(self):
        return self.category == BomCategory.BUNDLE

    @property
    def is_autogenerated(self):
        return self.designator.startswith(AUTOGENERATED_PREFIX)

    def __post_init__(self) -> None:
        if isinstance(self.category, str) and self.category.lower() == "bundle":
            self.category = BomCategory.BUNDLE
        else:
            self.category = BomCategory.CABLE
        self.notes = MultilineHypertext.to(self.notes)

        # TODO: style management should be separated from this logic...
        self.bgcolor = SingleColor(self.bgcolor)
        self.bgcolor_title = SingleColor(self.bgcolor_title)
        self.color = MultiColor(self.color)

        if isinstance(self.image, dict):
            self.image = Image(**self.image)

        # TODO:
        # allow gauge, length, and other fields to be lists too (like part numbers),
        # and assign them the same way to bundles.

        self.gauge = NumberAndUnit.to_number_and_unit(self.gauge, "mm2")
        self.length = NumberAndUnit.to_number_and_unit(self.length, "m")
        self.amount = self.length  # for BOM

        if self.wirecount:  # number of wires explicitly defined
            self.wirecount = int(self.wirecount)
            if self.colors:  # use custom color palette (partly or looped if needed)
                self.colors = [
                    self.colors[i % len(self.colors)] for i in range(self.wirecount)
                ]
            elif self.color_code:
                # use standard color palette (partly or looped if needed)
                if self.color_code not in COLOR_CODES:
                    raise Exception("Unknown color code")
                self.colors = [
                    get_color_by_colorcode_index(self.color_code, i)
                    for i in range(self.wirecount)
                ]
            elif self.color:
                self.colors = [self.color[w] for w in range(self.wirecount)]
            else:  # no colors defined, add dummy colors
                self.colors = [""] * self.wirecount

        else:  # wirecount implicit in length of color list
            if not self.colors:
                raise Exception(
                    "Unknown number of wires. "
                    "Must specify wirecount or colors (implicit length)"
                )
            self.wirecount = len(self.colors)

        if self.wirelabels:
            if self.shield and "s" in self.wirelabels:
                raise Exception(
                    '"s" may not be used as a wire label for a shielded cable.'
                )

        # if lists of part numbers are provided,
        # check this is a bundle and that it matches the wirecount.
        for idfield in [self.manufacturer, self.mpn, self.supplier, self.spn, self.pn]:
            if isinstance(idfield, list):
                if self.is_bundle:
                    # check the length
                    if len(idfield) != self.wirecount:
                        raise Exception("lists of part data must match wirecount")
                else:
                    raise Exception("lists of part data are only supported for bundles")

        # all checks have passed
        wire_tuples = zip_longest(
            # TODO: self.wire_ids
            self.colors,
            self.wirelabels,
        )
        for wire_index, (wire_color, wire_label) in enumerate(wire_tuples):
            id = wire_index + 1
            color = MultiColor(wire_color)[wire_index]
            by_idx = lambda x: x[wire_index] if isinstance(x, list) else x
            pn = by_idx(self.pn)
            manufacturer = by_idx(self.manufacturer)
            mpn = by_idx(self.mpn)
            supplier = by_idx(self.supplier)
            spn = by_idx(self.spn)

            self.wire_objects[id] = WireClass(
                pn=pn,
                manufacturer=manufacturer,
                mpn=mpn,
                supplier=supplier,
                spn=spn,
                parent=self.designator,
                # wire-specific properties
                index=wire_index,  # TODO: wire_id
                id=str(id),  # TODO: wire_id
                label=wire_label,
                color=color,
                # inheritable from parent cable
                type=self.type,
                subtype=self.subtype,
                gauge=self.gauge,
                length=self.length,
                ignore_in_bom=self.ignore_in_bom,
            )

        if self.shield:
            index_offset = len(self.wire_objects)
            # TODO: add support for multiple shields
            id = "s"
            self.wire_objects[id] = ShieldClass(
                index=index_offset,
                id=id,
                label="Shield",
                color=(
                    MultiColor(self.shield)
                    if isinstance(self.shield, str)
                    else MultiColor(None)
                ),
                parent=self.designator,
            )

        if self.show_name is None:
            self.show_name = not self.is_autogenerated

        for i, item in enumerate(self.additional_components):
            if isinstance(item, dict):
                self.additional_components[i] = Component(
                    **item, category=BomCategory.ADDITIONAL, parent=self.designator
                )

    def _connect(
        self,
        from_pin_obj: [PinClass],
        via_wire_id: str,
        to_pin_obj: [PinClass],
    ) -> None:
        via_wire_obj = self.wire_objects[via_wire_id]
        self._connections.append(Connection(from_pin_obj, via_wire_obj, to_pin_obj))

    def wire_ins(self, wire_id):
        return [
            str(c.from_)
            for c in self._connections
            if c.via.id == wire_id and c.from_ is not None
        ]

    def wire_ins_str(self, wire_id):
        return ", ".join(self.wire_ins(wire_id))

    def wire_outs(self, wire_id):
        return [
            str(c.to)
            for c in self._connections
            if c.via.id == wire_id and c.to is not None
        ]

    def wire_outs_str(self, wire_id):
        return ", ".join(self.wire_outs(wire_id))

    def compute_qty_multipliers(self):
        # do not run before all connections in harness have been made!
        total_length = sum(
            [
                wire.length.number if wire.length else 0
                for wire in self.wire_objects.values()
            ]
        )
        qty_multipliers_computed = {
            "WIRECOUNT": len(self.wire_objects),
            "TERMINATIONS": 999,  # TODO
            "LENGTH": self.length.number if self.length else 0,
            "TOTAL_LENGTH": total_length,
        }
        for subitem in self.additional_components:
            if isinstance(subitem.qty_multiplier, QtyMultiplierCable):
                computed_factor = qty_multipliers_computed[subitem.qty_multiplier.name]
                # inherit component's length unit if appropriate
                if subitem.qty_multiplier.name.upper() in ["LENGTH", "TOTAL_LENGTH"]:
                    if subitem.qty.unit is not None:
                        raise Exception(
                            f"No unit may be specified when using"
                            f"{subitem.qty_multiplier} as a multiplier"
                        )
                    subitem.qty = NumberAndUnit(
                        number=subitem.qty.number, unit=self.length.unit
                    )

            elif isinstance(subitem.qty_multiplier, QtyMultiplierConnector):
                raise Exception("Used a connector multiplier in a cable!")
            else:  # int or float
                computed_factor = subitem.qty_multiplier
            subitem._qty_multiplier_computed = computed_factor



# -*- coding: utf-8 -*-
# Please don't import anything in this file to avoid issues when it is imported in setup.py

__version__ = "0.1"

CMD_NAME = "filare"  # Lower case command and module name for branded CLI
APP_NAME = "Filare"  # Application name in texts meant to be human readable
APP_URL = "https://github.com/laurierloi/Filare"



from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field, root_validator, validator


class ConfigBaseModel(BaseModel):
    """Common base for YAML-facing configuration models."""

    class Config:
        extra = Extra.allow
        allow_mutation = True
        arbitrary_types_allowed = True


class PinConfig(ConfigBaseModel):
    id: Optional[str] = None
    label: Optional[str] = None
    color: Optional[Union[str, List[str]]] = None
    note: Optional[str] = None


class ConnectorConfig(ConfigBaseModel):
    designator: str
    pincount: Optional[int] = None
    pins: Optional[List[PinConfig]] = None
    pinlabels: Optional[List[str]] = None
    pincolors: Optional[Union[List[str], Dict[str, str]]] = None
    loops: Optional[List[Dict[str, Any]]] = None
    style: Optional[str] = None
    images: Optional[List[str]] = None
    notes: Optional[List[str]] = None

    @root_validator(pre=True)
    def _set_pincount_from_inputs(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("pincount") is None:
            pins = values.get("pins") or []
            pinlabels = values.get("pinlabels") or []
            if isinstance(pins, dict):
                pins = [pins]
            if isinstance(pinlabels, dict):
                pinlabels = list(pinlabels.values())
            if pins:
                values["pincount"] = len(pins)
            elif pinlabels:
                values["pincount"] = len(pinlabels)
        return values

    @validator("pins", pre=True)
    def _coerce_pins(cls, value: Any) -> Any:
        if value is None:
            return value
        if isinstance(value, dict):
            return [value]
        return value

    @validator("loops", pre=True)
    def _ensure_loop_list(cls, value: Any) -> Any:
        if value is None:
            return value
        if isinstance(value, dict):
            return [value]
        return list(value)

    @validator("pincount", pre=True, always=True)
    def _derive_pincount(
        cls, value: Optional[int], values: Dict[str, Any]
    ) -> Optional[int]:
        if value is not None:
            return value
        pins = values.get("pins")
        if pins:
            return len(pins)
        pinlabels = values.get("pinlabels")
        if pinlabels:
            return len(pinlabels)
        return value


class WireConfig(ConfigBaseModel):
    label: Optional[str] = None
    color: Optional[Union[str, List[str]]] = None
    gauge: Optional[Any] = None
    length: Optional[Any] = None
    shield: Optional[Any] = None
    notes: Optional[List[str]] = None


class CableConfig(ConfigBaseModel):
    designator: str
    wirecount: Optional[int] = None
    colors: Optional[List[str]] = None
    shields: Optional[List[Any]] = None
    length: Optional[Any] = None
    wires: Optional[List[WireConfig]] = None
    notes: Optional[List[str]] = None
    style: Optional[str] = None

    @validator("colors", pre=True)
    def _coerce_colors(cls, value: Any) -> Any:
        if value is None:
            return value
        if isinstance(value, str):
            return [value]
        return list(value)

    @validator("wirecount", always=True)
    def _derive_wirecount(
        cls, value: Optional[int], values: Dict[str, Any]
    ) -> Optional[int]:
        if value is not None:
            return value
        wires = values.get("wires")
        if wires:
            return len(wires)
        colors = values.get("colors")
        if colors:
            return len(colors)
        return value


class ConnectionConfig(ConfigBaseModel):
    endpoints: List[str]
    net: Optional[str] = None
    color: Optional[Union[str, List[str]]] = None
    wire: Optional[str] = None
    bundle: Optional[str] = None
    notes: Optional[List[str]] = None

    @validator("endpoints", pre=True)
    def _coerce_endpoints(cls, value: Any) -> List[str]:
        if isinstance(value, str):
            return [value]
        if isinstance(value, (tuple, set)):
            return list(value)
        return list(value)


class MetadataConfig(ConfigBaseModel):
    title: Optional[str] = None
    pn: Optional[str] = None
    company: Optional[str] = None
    address: Optional[str] = None
    output_dir: Optional[Any] = None
    output_name: Optional[str] = None
    files: List[str] = Field(default_factory=list)
    sheet_total: Optional[int] = None
    sheet_current: Optional[int] = None
    sheet_name: Optional[str] = None
    titlepage: Optional[Any] = None
    output_names: List[str] = Field(default_factory=list)
    authors: Dict[str, Any] = Field(default_factory=dict)
    revisions: Dict[str, Any] = Field(default_factory=dict)
    template: Optional[Dict[str, Any]] = None
    use_qty_multipliers: Optional[bool] = None
    multiplier_file_name: Optional[str] = None


class PageOptionsConfig(ConfigBaseModel):
    bgcolor: Optional[Any] = None
    bgcolor_connector: Optional[Any] = None
    bgcolor_overview: Optional[Any] = None
    connector_overview_style: Optional[str] = None
    overview_inherit_styles: Optional[bool] = None
    font: Optional[str] = None
    dpi: Optional[int] = None
    width_mm: Optional[float] = None
    margin_mm: Optional[float] = None
    formats: Optional[List[str]] = None

    @validator("formats", pre=True)
    def _coerce_formats(cls, value: Any) -> Any:
        if value is None:
            return value
        if isinstance(value, str):
            return [value]
        return list(value)



import textwrap

import pytest

from filare.cli import cli


@pytest.mark.parametrize("formats", ["tb"])
def test_shared_bom_contains_both_harnesses(tmp_path, formats):
    pn = "TEST-DOC"

    metadata_path = tmp_path / "metadata.yml"
    metadata_path.write_text(
        textwrap.dedent(
            f"""\
            metadata:
              pn: {pn}
              company: Test Company
              address: 123 Test Street
              authors: {{}}
              revisions: {{}}
              template:
                name: din-6771
                sheetsize: A4
            """
        )
    )

    harness_a = tmp_path / "harness_a.yml"
    harness_a.write_text(
        textwrap.dedent(
            """\
            connectors:
              J1:
                pincount: 2
              J2:
                pincount: 2

            cables:
              W1:
                wirecount: 2

            connections:
              -
                - J1: [1, 2]
                - W1: [1, 2]
                - J2: [1, 2]
            """
        )
    )

    harness_b = tmp_path / "harness_b.yml"
    harness_b.write_text(
        textwrap.dedent(
            """\
            connectors:
              P1:
                pincount: 3
              P2:
                pincount: 3

            cables:
              W2:
                wirecount: 3

            connections:
              -
                - P1: [1-3]
                - W2: [1-3]
                - P2: [1-3]
            """
        )
    )

    cli.callback(  # type: ignore[attr-defined]
        files=(harness_a, harness_b),
        formats=formats,
        components=(),
        metadata=(metadata_path,),
        output_dir=tmp_path,
        output_name=None,
        version=False,
        use_qty_multipliers=False,
        multiplier_file_name="quantity_multipliers.txt",
    )

    harness_a_bom = tmp_path / f"{harness_a.stem}.tsv"
    harness_b_bom = tmp_path / f"{harness_b.stem}.tsv"
    shared_bom = tmp_path / "shared_bom.tsv"

    for output_file in (harness_a_bom, harness_b_bom, shared_bom):
        assert output_file.exists(), f"Missing {output_file}"

    shared_bom_text = shared_bom.read_text()
    expected_harness_names = [
        f"{pn}-{harness_a.stem}",
        f"{pn}-{harness_b.stem}",
    ]
    for harness_name in expected_harness_names:
        assert harness_name in shared_bom_text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

from filare.flows import build_harness_from_files


def parse(
    inp: List[Path],
    metadata_files: List[Path],
    return_types: Union[None, str, Tuple[str]] = None,
    output_formats: Union[None, str, Tuple[str]] = None,
    output_dir: Path = None,
    extra_metadata: Dict = {},
    shared_bom: Dict = {},
    output_name_override: str = None,
    connector_view: str = "detailed",
    metadata_output_name: str = None,
    update_shared_bom: bool = True,
) -> Any:
    """
    Wrapper to build and optionally render a Harness from YAML inputs.
    """

    return build_harness_from_files(
        inp=inp,
        metadata_files=metadata_files,
        return_types=return_types,
        output_formats=output_formats,
        output_dir=output_dir,
        extra_metadata=extra_metadata,
        shared_bom=shared_bom,
        output_name_override=output_name_override,
        connector_view=connector_view,
        metadata_output_name=metadata_output_name,
        update_shared_bom=update_shared_bom,
    )



import logging
from dataclasses import asdict, dataclass, field, fields
from enum import Enum, IntEnum
from textwrap import dedent
from typing import List


@dataclass
class Notes:
    notes: List[str] = field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.notes, Notes):
            self.notes = self.notes.notes

    def __repr__(self):
        return self.as_html_list()

    def as_html_list(self):
        if self.notes:
            lines = "\n".join(f"  <li>{note}</li>" for note in self.notes)
            return f"<ol>\n{lines}\n</ol>"
        return ""


# TODO: nearly same method is within page_options.py, standardize?
def get_page_notes(parsed_data, page_name: str):
    """Get the page options

    uses: the page\'s notes   -> general notes -> default notes
        ('{page_name}_notes') ->    ('notes')  -> {}
    """
    page_notes_name = f"{page_name}_notes"
    notes = parsed_data.get(page_notes_name, parsed_data.get("notes", []))
    return Notes(notes=notes)



"""Flow that builds a Harness model from YAML inputs and renders outputs."""

import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from filare.models.metadata import Metadata
from filare.models.notes import Notes, get_page_notes
from filare.models.options import PageOptions, get_page_options
from filare.models.connector import ConnectorModel
from filare.models.cable import CableModel
from filare.models.component import ComponentModel
from filare.parser import parse_concat_merge_files
from filare.models.dataclasses import AUTOGENERATED_PREFIX
from filare.models.harness import Harness
from filare.models.utils import expand, get_single_key_and_value, smart_file_resolve

from .render_outputs import render_harness_outputs


def _build_metadata(yaml_file, yaml_data, extra_metadata, metadata_output_name):
    return Metadata(
        **{
            **{
                "output_name": metadata_output_name,
                "title": yaml_file.stem,
            },
            **yaml_data.get("metadata", {}),
            **extra_metadata,
        }
    )


def _collect_templates(yaml_data, image_paths):
    template_connectors = {}
    template_cables = {}
    connection_sets = []
    sections = ["connectors", "cables", "connections"]
    types = [dict, dict, list]
    for sec, ty in zip(sections, types):
        if sec in yaml_data and type(yaml_data[sec]) == ty:
            if len(yaml_data[sec]) > 0:
                if ty == dict:
                    for key, attribs in yaml_data[sec].items():
                        image = attribs.get("image")
                        if isinstance(image, dict):
                            image_path = Path(image["src"])
                            if image_path and not image_path.is_absolute():
                                attribs["image"]["src"] = smart_file_resolve(
                                    image_path, image_paths
                                )
                        if sec == "connectors":
                            template_connectors[key] = attribs
                        elif sec == "cables":
                            template_cables[key] = attribs
            else:
                pass
        else:
            if ty == dict:
                yaml_data[sec] = {}
            elif ty == list:
                yaml_data[sec] = []
    connection_sets = yaml_data["connections"]
    return template_connectors, template_cables, connection_sets


def build_harness_from_models(
    connector_models: List[Union[ConnectorModel, Dict[str, Any]]],
    cable_models: List[Union[CableModel, Dict[str, Any]]],
    metadata: Metadata,
    options: PageOptions,
    notes: Notes = Notes([]),
    additional_bom: Optional[List[Union[ComponentModel, Dict[str, Any]]]] = None,
) -> Harness:
    """Build a Harness from already-instantiated connector/cable models."""
    harness = Harness(metadata=metadata, options=options, notes=notes)
    for connector in connector_models:
        harness.add_connector_model(connector)
    for cable in cable_models:
        harness.add_cable_model(cable)
    for item in additional_bom or []:
        harness.add_additional_bom_item(item)
    harness.populate_bom()
    return harness


def _normalize_connection_set(
    connection_set,
    template_separator_char,
    designators_and_templates,
    autogenerated_designators,
):
    def resolve_designator(inp, separator):
        if separator in inp:
            if inp.count(separator) > 1:
                raise Exception(f"{inp} - Found more than one separator ({separator})")
            template, designator = inp.split(separator)
            if designator == "":
                autogenerated_designators[template] = (
                    autogenerated_designators.get(template, 0) + 1
                )
                designator = (
                    f"{AUTOGENERATED_PREFIX}"
                    f"{template}_{autogenerated_designators[template]}"
                )
            if designator in designators_and_templates:
                if designators_and_templates[designator] != template:
                    raise Exception(
                        f"Trying to redefine {designator}"
                        f" from {designators_and_templates[designator]} to {template}"
                    )
            else:
                designators_and_templates[designator] = template
        else:
            template, designator = (inp, inp)
            if designator in designators_and_templates:
                pass
            else:
                designators_and_templates[designator] = template
        return (template, designator)

    connectioncount = []
    for entry in connection_set:
        if isinstance(entry, list):
            connectioncount.append(len(entry))
        elif isinstance(entry, dict):
            connectioncount.append(len(expand(list(entry.values())[0])))

    if not any(connectioncount):
        raise ValueError("No connection count found in connection set")

    if len(set(connectioncount)) > 1:
        pretty_sets = [
            f'length {len(v)}: "{k}": {v}' for s in connection_set for k, v in s.items()
        ]
        prettyer_sets = "\n\t" + "\n\t".join(pretty_sets)
        raise Exception(
            f"All items in connection set must reference the same number of connections\nIt is not the case for:{prettyer_sets}"
        )

    connectioncount = connectioncount[0]

    for index, entry in enumerate(connection_set):
        if isinstance(entry, str):
            connection_set[index] = [entry] * connectioncount

    for index, entry in enumerate(connection_set):
        if isinstance(entry, list):
            for subindex, item in enumerate(entry):
                template, designator = resolve_designator(item, template_separator_char)
                connection_set[index][subindex] = designator
        elif isinstance(entry, dict):
            key = list(entry.keys())[0]
            template, designator = resolve_designator(key, template_separator_char)
            value = entry[key]
            connection_set[index] = {designator: value}

    for index, entry in enumerate(connection_set):
        if isinstance(entry, list):
            connection_set[index] = [{designator: 1} for designator in entry]
        elif isinstance(entry, dict):
            designator = list(entry.keys())[0]
            pinlist = expand(entry[designator])
            connection_set[index] = [{designator: pin} for pin in pinlist]

    return connection_set, connectioncount


def build_harness_from_files(
    inp: List[Path],
    metadata_files: List[Path],
    return_types: Union[None, str, Tuple[str]] = None,
    output_formats: Union[None, str, Tuple[str]] = None,
    output_dir: Path = None,
    extra_metadata: Dict = {},
    shared_bom: Dict = {},
    output_name_override: str = None,
    connector_view: str = "detailed",
    metadata_output_name: str = None,
    update_shared_bom: bool = True,
) -> Any:
    if not output_formats and not return_types:
        raise Exception("No output formats or return types specified")

    yaml_file = inp[-1]
    yaml_data = parse_concat_merge_files(inp, metadata_files)

    image_paths = {f.parent for f in inp if f.parent.is_dir()}

    output_dir = yaml_file.parent if not output_dir else output_dir
    output_name = output_name_override if output_name_override else yaml_file.stem
    metadata_output_name = metadata_output_name or output_name

    try:
        metadata = _build_metadata(
            yaml_file, yaml_data, extra_metadata, metadata_output_name
        )
    except TypeError:
        logging.error(
            "Metadata definition is missing an argument, refer to trace for which one\n\tsee src/filaree/metdata.py for a definition of the metadata fields"
        )
        raise

    harness = Harness(
        metadata=metadata,
        options=get_page_options(yaml_data, output_name),
        notes=get_page_notes(yaml_data, output_name),
        shared_bom=shared_bom,
    )
    designators_and_templates = {}
    autogenerated_designators = {}

    template_connectors, template_cables, connection_sets = _collect_templates(
        yaml_data, image_paths
    )

    alternating_types = ["connector", "cable"]
    expected_type = None

    def check_type(designator, template, actual_type):
        nonlocal expected_type
        if not expected_type:
            expected_type = actual_type

        if actual_type != expected_type:
            raise Exception(
                f'Expected {expected_type}, but "{designator}" ("{template}") is {actual_type}'
            )

    def alternate_type():
        nonlocal expected_type
        expected_type = alternating_types[1 - alternating_types.index(expected_type)]

    template_separator_char = harness.options.template_separator

    for connection_set in connection_sets:
        connection_set, connectioncount = _normalize_connection_set(
            connection_set,
            template_separator_char,
            designators_and_templates,
            autogenerated_designators,
        )

        expected_type = None

        for entry in connection_set:
            for item in entry:
                designator = list(item.keys())[0]
                template = designators_and_templates[designator]

                if designator in harness.connectors:
                    check_type(designator, template, "connector")
                elif template in template_connectors.keys():
                    check_type(designator, template, "connector")
                    harness.add_connector(
                        designator=designator, **template_connectors[template]
                    )

                elif designator in harness.cables:
                    check_type(designator, template, "cable")
                elif template in template_cables.keys():
                    check_type(designator, template, "cable")
                    harness.add_cable(
                        designator=designator, **template_cables[template]
                    )
                else:
                    raise Exception(f"{template} is an unknown template/designator")

            alternate_type()

        connection_set = list(map(list, zip(*connection_set)))

        for index_entry, entry in enumerate(connection_set):
            for index_item, item in enumerate(entry):
                designator = list(item.keys())[0]

                if designator in harness.cables:
                    if index_item == 0:
                        from_name, from_pin = (None, None)
                    else:
                        from_name, from_pin = get_single_key_and_value(
                            entry[index_item - 1]
                        )
                    via_name, via_pin = (designator, item[designator])
                    if index_item == len(entry) - 1:
                        to_name, to_pin = (None, None)
                    else:
                        to_name, to_pin = get_single_key_and_value(
                            entry[index_item + 1]
                        )
                    harness.connect(
                        from_name, from_pin, via_name, via_pin, to_name, to_pin
                    )

    if "additional_bom_items" in yaml_data:
        for line in yaml_data["additional_bom_items"]:
            try:
                harness.add_additional_bom_item(line)
            except TypeError as e:
                logging.error(f"Failed to add line {line} as an additional bom item")
                raise

    if connector_view == "overview":
        harness.orient_connectors_overview()

    if connector_view == "overview":
        for connector in harness.connectors.values():
            if connector.style != "simple":
                connector.style = "overview"
        for cable in harness.cables.values():
            cable.style = "overview"

    harness.populate_bom()

    if output_formats:
        render_harness_outputs(harness, output_dir, output_name, output_formats)

    if return_types:
        if isinstance(return_types, str):
            return_types = [return_types]

        return_types = [t.lower() for t in return_types]

        returns = {}
        for rt in return_types:
            if rt == "png":
                returns["png"] = harness.png
            if rt == "svg":
                returns["svg"] = harness.svg
            if rt == "harness":
                returns["harness"] = harness
            if rt == "shared_bom":
                returns["shared_bom"] = harness.shared_bom

        return returns



from typing import ClassVar, Dict, List, Union

import tabulate as tabulate_module
from pydantic import BaseModel, Field, root_validator

from filare.models.numbers import NumberAndUnit
from filare.models.partnumber import PartNumberInfo
from filare.models.utils import remove_links
from filare.render.templates import get_template


class BomEntryBase(BaseModel):
    qty: NumberAndUnit
    partnumbers: PartNumberInfo
    id: str = ""
    amount: Union[NumberAndUnit, None] = None
    qty_multiplier: Union[int, float, object] = 1
    description: str = ""
    category: str = ""
    ignore_in_bom: bool = False
    designators: List = Field(default_factory=list)
    per_harness: Dict = Field(default_factory=dict)
    MAX_PRINTED_DESCRIPTION: ClassVar[int] = 40
    MAX_PRINTED_DESIGNATORS: ClassVar[int] = 2
    restrict_printed_lengths: bool = True
    scaled_per_harness: bool = False
    BOM_KEY_TO_COLUMNS: ClassVar[Dict[str, str]] = {
        "id": "#",
        "qty": "Qty",
        "unit": "Unit",
        "description": "Description",
        "designators": "Designators",
        "per_harness": "Per Harness",
    }

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = True

    @root_validator
    def _scale_qty(cls, values):
        try:
            values["qty_multiplier"] = float(values.get("qty_multiplier", 1))
        except Exception:
            values["qty_multiplier"] = 1
        if values.get("qty") is not None:
            values["qty"] *= float(values.get("qty_multiplier", 1))
        if values.get("amount") is not None:
            values["amount"] *= float(values.get("qty_multiplier", 1))
        return values

    def __hash__(self):
        return hash((self.partnumbers, self.description))

    def __eq__(self, other):
        if isinstance(other, list):
            return False
        elif isinstance(other, BomEntryBase):
            return hash(self) == hash(other)

    def __add__(self, other):
        if isinstance(other, list):
            return [self + o for o in other]
        elif isinstance(other, BomEntryBase):
            self.qty += other.qty
            self.designators += other.designators
            return self
        else:
            raise NotImplementedError(f"__add__ for {type(other)}")

    def scale_per_harness(self, multipliers):
        if self.scaled_per_harness:
            logging.warning("scale_per_harness() was called twice for item with no ID")
            return
        for k, v in self.per_harness.items():
            if k in multipliers:
                v["qty"] *= multipliers[k]
        self.scaled_per_harness = True

    @property
    def unit(self):
        return self.qty.unit if self.qty.unit else ""

    @property
    def description_clean(self):
        desc = self.description
        if self.restrict_printed_lengths:
            desc = (
                self.description[: self.MAX_PRINTED_DESCRIPTION] + "..."
                if len(self.description) > self.MAX_PRINTED_DESCRIPTION
                else self.description
            )
        return remove_links(desc)

    @property
    def designators_str(self):
        if self.restrict_printed_lengths and len(self.designators) > 0:
            designators = self.designators[: self.MAX_PRINTED_DESIGNATORS]
            more = len(self.designators) - len(designators)
            return ", ".join(designators) + (" (...)" if more > 0 else "")
        else:
            return ", ".join(self.designators)

    @property
    def per_harness_str(self):
        if not self.per_harness:
            return ""
        parts = []
        for harness, data in self.per_harness.items():
            qty = data["qty"]
            parts.append(f"{harness}: {qty}")
        return "; ".join(parts)

    def as_list(self, filter_empty=False, include_per_harness=False):
        lst = [
            getattr(self, "id", ""),
            self.qty.number,
            self.unit,
            self.description_clean,
            self.designators_str,
        ]
        if include_per_harness:
            lst.append(self.per_harness_str)
        if not filter_empty:
            return lst
        return [item for item in lst if item]


class BomEntry(BomEntryBase):
    id: str = ""
    BOM_KEY_TO_COLUMNS: ClassVar[Dict[str, str]] = (
        BomEntryBase.BOM_KEY_TO_COLUMNS.copy()
    )

    def __repr__(self):
        return f"{self.id}: {self.partnumbers}, {self.qty}"

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = True


class BomRender:
    def __init__(self, header, rows, strip_empty_columns=False, columns_class=None):
        self.header = header
        self.rows = rows
        self.strip_empty_columns = strip_empty_columns
        self.options = None
        self.content = rows
        if columns_class is None:
            self.columns_class = [""] * len(header)
        else:
            self.columns_class = columns_class

    def as_tsv(self):
        tsv = tabulate_module.tabulate(self.rows, self.header, tablefmt="tsv")
        return tsv

    @property
    def headers(self):
        return self.header

    def render(self, page_options=None, bom_options=None):
        if page_options is None:
            page_options = {}
        if bom_options is None:
            bom_options = BomRenderOptions()
        if self.strip_empty_columns:
            # remove empty columns in header and rows
            index_to_remove = [i for i, v in enumerate(self.header) if v == ""]
            stripped_rows = self.rows.copy()
            stripped_header = self.header.copy()
            for i in index_to_remove[::-1]:
                stripped_header.pop(i)
                for row in stripped_rows:
                    row.pop(i)

            header = stripped_header
            rows = stripped_rows
        else:
            header = self.header
            rows = self.rows

        self.options = bom_options
        return get_template("bom.html").render(
            {
                "bom": self,
                "table_header": header,
                "table_rows": rows,
                "options": page_options,
            }
        )


class BomRenderOptions:
    def __init__(
        self,
        restrict_printed_lengths=False,
        filter_entries=True,
        no_per_harness=False,
        reverse=False,
    ):
        self.restrict_printed_lengths = restrict_printed_lengths
        self.filter_entries = filter_entries
        self.no_per_harness = no_per_harness
        self.reverse = reverse


class BomContent(dict):
    def __init__(self, bom):
        self._bom = bom
        super().__init__(bom)

    def filter_entries(self, restrict_printed_lengths=True, filter_entries=True):
        # TODO: Refactor to not mutate in place
        for _, entry in self.items():
            entry.restrict_printed_lengths = restrict_printed_lengths

        # remove entries without partnumbers or quantities
        if filter_entries:
            _keys_to_remove = []
            for key, entry in self.items():
                if not entry.partnumbers or entry.qty.number <= 0:
                    _keys_to_remove.append(key)
            for key in _keys_to_remove:
                del self[key]

    def get_bom_render(self, options=BomRenderOptions()):
        """Return a BomRender object using class properties for BOM rendering options."""

        self.filter_entries(
            options.restrict_printed_lengths,
            options.filter_entries,
        )

        header = [
            BomEntry.BOM_KEY_TO_COLUMNS["id"],
            BomEntry.BOM_KEY_TO_COLUMNS["qty"],
            BomEntry.BOM_KEY_TO_COLUMNS["unit"],
            BomEntry.BOM_KEY_TO_COLUMNS["description"],
            BomEntry.BOM_KEY_TO_COLUMNS["designators"],
        ]
        include_per_harness = not getattr(options, "no_per_harness", False)
        if include_per_harness:
            header.append(BomEntry.BOM_KEY_TO_COLUMNS["per_harness"])
        rows = []
        columns_class = [
            "bom_col_id",
            "bom_col_qty",
            "bom_col_unit",
            "bom_col_description",
            "bom_col_designators",
        ]
        if include_per_harness:
            columns_class.append("bom_col_per_harness")

        # TODO: in reverse mode we probably want to re-sort BOM by description here

        for entry in self.values():
            row = entry.as_list(False, include_per_harness=include_per_harness)
            rows.append(row)
        if options.reverse:
            rows = rows[::-1]
        return BomRender(header, rows, columns_class=columns_class)


def print_bom_table(bom):
    header = [
        BomEntry.BOM_KEY_TO_COLUMNS["id"],
        BomEntry.BOM_KEY_TO_COLUMNS["qty"],
        BomEntry.BOM_KEY_TO_COLUMNS["unit"],
        BomEntry.BOM_KEY_TO_COLUMNS["description"],
        BomEntry.BOM_KEY_TO_COLUMNS["designators"],
    ]

    rows = []
    for entry in bom.values():
        rows.append(entry.as_list())

    print(tabulate_module.tabulate(rows, header))



"""Centralized settings loaded from environment with WV_ prefix."""

import os
from typing import Optional

from pydantic import BaseSettings, Field


class FilareSettings(BaseSettings):
    graphviz_engine: Optional[str] = Field(
        default=None, env="WV_GRAPHVIZ_ENGINE"
    )  # e.g., dot, neato
    debug: bool = Field(default=False, env="WV_DEBUG")

    class Config:
        env_prefix = "WV_"
        case_sensitive = False


settings = FilareSettings()



"""Flow-level orchestration helpers."""

from filare.flows.build_harness import build_harness_from_files
from filare.flows.render_outputs import render_harness_outputs

__all__ = ["build_harness_from_files", "render_harness_outputs"]
